"""
This module contains custom errors used in the Firedust SDK.
"""


# AUTHENTICATION ERRORS
class MissingFiredustKeyError(Exception):
    def __init__(self) -> None:
        message = """
        FIREDUST_API_KEY environment variable is not found. 
        Get your API key at https://www.firedust.dev
        """
        super().__init__(message)


# API ERRORS
class APIError(Exception):
    def __init__(self, message: str, code: int) -> None:
        self.code = code
        self.message = message
        super().__init__(message)

    def __str__(self) -> str:
        return f"APIError (code={self.code}): {super().__str__()}"


# ASSISTANT ERRORS
class AssistantError(Exception):
    """
    Default error for the Assistant class.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)


# MEMORY ERRORS
class MemoryError(Exception):
    """
    Default error for the Memory class.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)


# INTERFACE ERRORS
class SlackError(Exception):
    """
    Default error for the Slack and SlackInterface class.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)



from datetime import datetime
from typing import Literal, Optional
from uuid import UUID

from pydantic import Field, field_serializer, field_validator

from firedust.utils import checks

from .base import UNIX_TIMESTAMP, BaseConfig

MAX_MEMORY_CONTENT: int = 4000  # max characters for the context of a memory item


class MemoryItem(BaseConfig):
    """
    A memory used by the AI assistant. Memories are data items that are stored in the assistant's database
    and, when relevant, used as context to answer questions and perform tasks.

    Args:
        assistant (str): Assistant name.
        content (str): The content of the memory.
        embedding (Sequence[float]): The embedding of the memory.
        timestamp (UNIX_TIMESTAMP): The time when the memory was created.
        type (Literal["text", "image", "audio", "video"]): The type of the memory. Defaults to "text".
        source (str, optional): The source of the memory. Defaults to None.
        relevance (float, optional): Relevance score, provided by the vector search.
    """

    assistant: str = Field(...)
    content: str = Field(..., max_length=MAX_MEMORY_CONTENT)
    timestamp: UNIX_TIMESTAMP = Field(
        default_factory=lambda: datetime.now().timestamp()
    )
    type: Literal["text", "image", "audio", "video"] = Field(default="text")
    source: Optional[str] = Field(default=None)
    relevance: Optional[float] = Field(default=None)

    @field_validator("timestamp")
    @classmethod
    def validate_timestamp(cls, timestamp: UNIX_TIMESTAMP) -> UNIX_TIMESTAMP:
        # Check that the timestamp is UNIX format
        try:
            checks.is_unix_timestamp(timestamp)
        except ValueError as e:
            raise ValueError(f"Invalid timestamp: {e}")
        return timestamp

    @field_serializer("assistant", when_used="always")
    def serialize_collection_id(self, value: UUID) -> str:
        return str(value)



from typing import Literal, Optional

from pydantic import BaseModel


class InterfaceConfig(BaseModel):
    """
    Represents an interface of the assistant.
    """

    interface: Literal["slack"]  # space for more interfaces


class SlackTokens(BaseModel):
    """
    Represents the credentials of a Slack app.
    """

    app_token: str
    bot_token: str


class SlackCredentials(BaseModel):
    """
    Represents the credentials of a Slack app.
    """

    signing_secret: str
    client_id: str
    client_secret: str


class SlackConfig(InterfaceConfig):
    """
    Configuration for Slack InterfaceConfig.
    """

    description: str
    greeting: str  # A message that the assistant will greet users with.
    interface: Literal["slack"] = "slack"
    app_id: Optional[str] = None  # This is filled-in by the API
    credentials: Optional[SlackCredentials] = None  # This is filled-in by the API
    tokens: Optional[SlackTokens] = None  # This is filled-in by the API


class Interfaces(BaseModel):
    """
    Represents the interfaces of the assistant.
    """

    slack: Optional[SlackConfig] = None
    # add more interfaces as they come along



from firedust._assistant.interface.slack import AsyncSlackInterface, SlackInterface
from firedust.types import AssistantConfig
from firedust.utils.api import AsyncAPIClient, SyncAPIClient


class Interface:
    """
    A collection of methods to to deploy and interact with the assistant
    on various interfaces.

    Examples:
    - https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack.py
    -
    """

    def __init__(self, config: AssistantConfig, api_client: SyncAPIClient) -> None:
        # configuration
        self.config = config
        self.api_client = api_client

        # interfaces
        self.slack = SlackInterface(self.config, self.api_client)


class AsyncInterface:
    """
    A collection of async methods to to deploy and interact with the assistant
    on various interfaces asynchronously.

    Examples:
    - https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack.py
    -
    """

    def __init__(self, config: AssistantConfig, api_client: AsyncAPIClient) -> None:
        # configuration
        self.config = config
        self.api_client = api_client

        # interfaces
        self.slack = AsyncSlackInterface(self.config, self.api_client)



import os
from types import TracebackType
from typing import Any, AsyncIterator, Dict, Iterator, Optional, Type

import httpx

from firedust.utils.errors import MissingFiredustKeyError

BASE_URL = "http://localhost:3002"
# BASE_URL = "https://api.firedust.dev"
TIMEOUT = 300


class BaseAPIClient:
    def __init__(self, api_key: Optional[str] = None, base_url: str = BASE_URL) -> None:
        api_key = api_key or os.environ.get("FIREDUST_API_KEY")
        if not api_key:
            raise MissingFiredustKeyError()

        self.base_url = base_url
        self.api_key = api_key
        self.headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}",
        }


class SyncAPIClient(BaseAPIClient):
    def __init__(self, api_key: Optional[str] = None, base_url: str = BASE_URL) -> None:
        super().__init__(api_key, base_url)
        self.client: httpx.Client = httpx.Client(timeout=TIMEOUT, headers=self.headers)

    def __enter__(self) -> "SyncAPIClient":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.client.close()

    def __del__(self) -> None:
        self.client.close()

    def get(self, url: str, params: Optional[Dict[str, Any]] = None) -> httpx.Response:
        return self._request("get", url, params=params)

    def post(self, url: str, data: Optional[Dict[str, Any]] = None) -> httpx.Response:
        return self._request("post", url, data=data)

    def put(self, url: str, data: Optional[Dict[str, Any]] = None) -> httpx.Response:
        return self._request("put", url, data=data)

    def delete(
        self, url: str, params: Optional[Dict[str, Any]] = None
    ) -> httpx.Response:
        return self._request("delete", url, params=params)

    def get_stream(
        self, url: str, params: Optional[Dict[str, Any]] = None
    ) -> Iterator[bytes]:
        url = self.base_url + url
        with self.client.stream("get", url, params=params) as response:
            for chunk in response.iter_bytes():
                yield chunk

    def post_stream(
        self, url: str, data: Optional[Dict[str, Any]] = None
    ) -> Iterator[bytes]:
        url = self.base_url + url
        with self.client.stream("post", url, json=data) as response:
            for chunk in response.iter_bytes():
                yield chunk

    def _request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> httpx.Response:
        url = self.base_url + url
        response = self.client.request(method, url, params=params, json=data)
        return response

    def close(self) -> None:
        """
        Close the underlying HTTP client.
        """
        self.client.close()


class AsyncAPIClient(BaseAPIClient):
    def __init__(self, api_key: Optional[str] = None, base_url: str = BASE_URL) -> None:
        super().__init__(api_key, base_url)
        self.client = httpx.AsyncClient(timeout=TIMEOUT, headers=self.headers)

    async def __aenter__(self) -> "AsyncAPIClient":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.client.aclose()

    async def __del__(self) -> None:
        await self.client.aclose()

    async def get(
        self, url: str, params: Optional[Dict[str, Any]] = None
    ) -> httpx.Response:
        return await self._request("get", url, params=params)

    async def post(
        self, url: str, data: Optional[Dict[str, Any]] = None
    ) -> httpx.Response:
        return await self._request("post", url, data=data)

    async def put(
        self, url: str, data: Optional[Dict[str, Any]] = None
    ) -> httpx.Response:
        return await self._request("put", url, data=data)

    async def delete(
        self, url: str, params: Optional[Dict[str, Any]] = None
    ) -> httpx.Response:
        return await self._request("delete", url, params=params)

    async def get_stream(
        self, url: str, params: Optional[Dict[str, Any]] = None
    ) -> AsyncIterator[bytes]:
        url = self.base_url + url
        async with self.client.stream("get", url, params=params) as response:
            async for chunk in response.aiter_bytes():
                yield chunk

    async def post_stream(
        self, url: str, data: Optional[Dict[str, Any]] = None
    ) -> AsyncIterator[bytes]:
        url = self.base_url + url
        async with self.client.stream("post", url, json=data) as response:
            async for chunk in response.aiter_bytes():
                yield chunk

    async def _request(
        self,
        method: str,
        url: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
    ) -> httpx.Response:
        url = self.base_url + url
        response = await self.client.request(method, url, params=params, json=data)
        return response

    async def close(self) -> None:
        """
        Close the underlying HTTP client.
        """
        await self.client.aclose()



from typing import List

from firedust.utils.api import SyncAPIClient
from firedust.utils.errors import APIError


def text(content: str) -> List[float]:
    """
    Generates embeddings for the given string.

    Args:
        content (str): The text to generate embeddings for.

    Returns:
        List[float]: The embeddings for the text.
    """
    api_client = SyncAPIClient()
    response = api_client.post(
        "/data/embed/text",
        {
            "content": content,
        },
    )
    if not response.is_success:
        raise APIError(
            code=response.status_code,
            message=f"Failed to generate embeddings: {response.text}",
        )
    embeddings: List[float] = response.json()["data"]["embedding"]
    return embeddings



from typing import List

from pydantic import BaseModel, field_validator

from .base import INFERENCE_MODEL
from .interface import Interfaces

ASSISTANT_NAME = str


class AssistantConfig(BaseModel):
    """
    The configuration of an AI Assistant.

    Args:
        name (str): The name of the assistant.
        instructions (str): The instructions of the assistant.
        model (INFERENCE_MODEL, optional): The inference model used by the assistant. Defaults to "openai/gpt-4".
        attached_memories (List[ASSISTANT_NAME], optional): A list of assistant names whose memories are shared with the current assistant. Defaults to [].
        interfaces (Sequence[Interface], optional): The deployments of the assistant. Defaults to None.
    """

    name: str
    instructions: str
    model: INFERENCE_MODEL = "openai/gpt-4"
    attached_memories: List[ASSISTANT_NAME] = []
    interfaces: Interfaces = Interfaces()

    @field_validator("name")
    @classmethod
    def validate_name(cls, name: str) -> str:
        if len(name) < 1:
            raise ValueError("Assistant name must be at least 1 character")
        return name



import logging

LOG: logging.Logger = logging.getLogger("firedust")


def configure_logger() -> None:
    LOG.setLevel(logging.DEBUG)

    # Create a console handler
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    # Create a formatter and add it to the handlers
    formatter = logging.Formatter(
        "[%(asctime)s - %(name)s - %(levelname)s] - %(message)s"
    )
    console_handler.setFormatter(formatter)

    # Add the handlers to the LOG
    LOG.addHandler(console_handler)



from typing import Literal, Optional

from firedust.utils.api import AsyncAPIClient, SyncAPIClient
from firedust.utils.errors import APIError


def fast(
    text: str,
    instructions: str,
    score_min: float = 0.0,
    score_max: float = 1.0,
    distribution: Literal["uniform", "normal", "exponential"] = "uniform",
    good_example: Optional[str] = None,
    bad_example: Optional[str] = None,
) -> float:
    """
    Scores the given text using the Firedust API.

    Args:
        text (str): The text to be scored.
        instructions (str): The instructions for scoring the text.
        score_min (float, optional): The minimum score. Defaults to 0.0.
        score_max (float, optional): The maximum score. Defaults to 1.0.
        distribution (Literal["uniform", "normal", "exponential"], optional): The distribution of the scores. Defaults to "uniform".
        good_example (str, optional): A good example of the text. Defaults to None.
        bad_example (str, optional): A bad example of the text. Defaults to None.

    Returns:
        float: The score of the text.
    """
    api_client = SyncAPIClient()
    response = api_client.post(
        "/data/score/fast",
        {
            "text": text,
            "instructions": instructions,
            "score_min": score_min,
            "score_max": score_max,
            "distribution": distribution,
            "good_example": good_example,
            "bad_example": bad_example,
        },
    )
    if not response.is_success:
        raise APIError(
            code=response.status_code,
            message=f"Failed to score text: {response.text}",
        )
    return float(response.json()["data"]["score"])


async def fast_async(
    text: str,
    instructions: str,
    score_min: float = 0.0,
    score_max: float = 1.0,
    distribution: Literal["uniform", "normal", "exponential"] = "uniform",
    good_example: Optional[str] = None,
    bad_example: Optional[str] = None,
) -> float:
    """
    Scores the given text using the Firedust API asynchronously.

    Args:
        text (str): The text to be scored.
        instructions (str): The instructions for scoring the text.
        score_min (float, optional): The minimum score. Defaults to 0.0.
        score_max (float, optional): The maximum score. Defaults to 1.0.
        distribution (Literal["uniform", "normal", "exponential"], optional): The distribution of the scores. Defaults to "uniform".
        good_example (str, optional): A good example of the text. Defaults to None.
        bad_example (str, optional): A bad example of the text. Defaults to None.

    Returns:
        float: The score of the text.
    """
    api_client = AsyncAPIClient()
    response = await api_client.post(
        "/data/score/fast",
        {
            "text": text,
            "instructions": instructions,
            "score_min": score_min,
            "score_max": score_max,
            "distribution": distribution,
            "good_example": good_example,
            "bad_example": bad_example,
        },
    )
    if not response.is_success:
        raise APIError(
            code=response.status_code,
            message=f"Failed to score text: {response.text}",
        )
    return float(response.json()["data"]["score"])



from pathlib import Path
from typing import List, Union
from uuid import UUID

from firedust.types import AssistantConfig
from firedust.utils.api import AsyncAPIClient, SyncAPIClient
from firedust.utils.errors import APIError


class Learning:
    """
    A collection of methods to add various types of data to the assistant's
    knowledge graph. The data is embedded into the assistant's memory and
    used when a relevant taskor question is asked.
    """

    def __init__(self, config: AssistantConfig, api_client: SyncAPIClient) -> None:
        self.assistant = config
        self.api_client = api_client

    def fast(self, text: str) -> List[UUID]:
        """
        The fastest way to add data to the assistant's memory. It becomes
        available real-time for the assistant to use when answering questions
        or performing tasks.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        assistant.learn.fast("The quick brown fox jumps over the lazy dog.")

        response = assistant.chat.message("Who jumps over the lazy dog?")
        assert "brown fox" in response.message.lower()
        ```

        Args:
            text (str): The text to learn.

        Returns:
            memory_ids (List[UUID]): A list of memory ids for the memory items that were created.
        """
        response = self.api_client.post(
            "/assistant/learn/fast",
            data={"assistant": self.assistant.name, "text": text},
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to teach the assistant: {response.text}",
            )

        memory_ids = [UUID(_id) for _id in response.json()["data"]]
        return memory_ids

    def pdf(self, pdf: Union[str, Path]) -> None:
        """
        Learn the content of a PDF file.

        Args:
            pdf (Union[str, Path]): The path to the PDF file.
        """
        raise NotImplementedError()
        # with open(pdf, "rb") as f:
        #     response = self.api_client.post(
        #         "/assistant/learn/pdf",
        #         data={"assistant": self.assistant.name, "pdf": f},
        #     )
        #     if not response.is_success:
        #         raise APIError(f"Failed to teach the assistant: {response.text}")

    def url(self, url: str) -> None:
        """
        Learn the content of a web page.

        Args:
            url (str): The URL to the resource.
        """
        raise NotImplementedError()
        # response = self.api_client.post(
        #     "/assistant/learn/url",
        #     data={"assistant": self.assistant.name, "url": url},
        # )
        # if not response.is_success:
        #     raise APIError(f"Failed to teach the assistant: {response.text}")

    def image(self, image: Union[str, Path]) -> None:
        """
        Learn the content of an image.

        Args:
            image (Union[str, Path]): The path to the image.
        """
        raise NotImplementedError()
        # with open(image, "rb") as f:
        #     response = self.api_client.post(
        #         "/assistant/learn/image",
        #         data={"assistant": self.assistant.name, "image": f},
        #     )
        #     if not response.is_success:
        #         raise APIError(f"Failed to teach the assistant: {response.text}")

    def audio(self, audio: Union[str, Path]) -> None:
        """
        Learn the content of an audio file.

        Args:
            audio (Union[str, Path]): The path to the audio file.
        """
        raise NotImplementedError()
        # with open(audio, "rb") as f:
        #     response = self.api_client.post(
        #         "/assistant/learn/audio",
        #         data={"assistant": self.assistant.name, "audio": f},
        #     )
        #     if not response.is_success:
        #         raise APIError(f"Failed to teach the assistant: {response.text}")

    def video(self, video: Union[str, Path]) -> None:
        """
        Learn the content of a video file.

        Args:
            video (Union[str, Path]): The path to the video file.
        """
        raise NotImplementedError()


class AsyncLearning:
    """
    A collection of async methods add various types of data to the assistant's
    knowledge graph. The data is embedded into the assistant's memory and
    is used when a relevant taskor question is asked.
    """

    def __init__(self, assistant: AssistantConfig, api_client: AsyncAPIClient) -> None:
        self.assistant = assistant
        self.api_client = api_client

    async def fast(self, text: str) -> List[UUID]:
        """
        The fastest way to add data to the assistant's memory. It becomes available
        real-time for the assistant to use when answering questions or performing tasks.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
            memory_ids = await assistant.learn.fast("The quick brown fox jumps over the lazy dog.")

            response = await assistant.chat.message("Who jumps over the lazy dog?")
            assert "brown fox" in response.message.lower()

        asyncio.run(main())
        ```

        Args:
            text (str): The text to learn.

        Returns:
            memory_ids (List[UUID]): A list of memory ids.
        """
        response = await self.api_client.post(
            "/assistant/learn/fast",
            data={"assistant": self.assistant.name, "text": text},
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to teach the assistant: {response.text}",
            )

        memory_ids = [UUID(_id) for _id in response.json()["data"]]
        return memory_ids

    async def pdf(self, pdf: Union[str, Path]) -> None:
        """
        Learn the content of a PDF file.

        Args:
            pdf (Union[str, Path]): The path to the PDF file.
        """
        raise NotImplementedError()
        # with open(pdf, "rb") as f:
        #     response = await self.api_client.post(
        #         "/assistant/learn/pdf",
        #         data={"assistant": self.assistant.name, "pdf": f},
        #     )
        #     if not response.is_success:
        #         raise APIError(f"Failed to teach the assistant: {response.text}")

    async def url(self, url: str) -> None:
        """
        Learn the content of a web page.

        Args:
            url (str): The URL to the resource.
        """
        raise NotImplementedError()
        # response = await self.api_client.post(
        #     "/assistant/learn/url",
        #     data={"assistant": self.assistant.name, "url": url},
        # )
        # if not response.is_success:
        #     raise APIError(f"Failed to teach the assistant: {response.text}")

    async def image(self, image: Union[str, Path]) -> None:
        """
        Learn the content of an image.

        Args:
            image (Union[str, Path]): The path to the image.
        """
        raise NotImplementedError()
        # with open(image, "rb") as f:
        #     response = await self.api_client.post(
        #         "/assistant/learn/image",
        #         data={"assistant": self.assistant.name, "image": f},
        #     )
        #     if not response.is_success:
        #         raise APIError(f"Failed to teach the assistant: {response.text}")

    async def audio(self, audio: Union[str, Path]) -> None:
        """
        Learn the content of an audio file.

        Args:
            audio (Union[str, Path]): The path to the audio file.
        """
        raise NotImplementedError()
        # with open(audio, "rb") as f:
        #     response = await self.api_client.post(
        #         "/assistant/learn/audio",
        #         data={"assistant": self.assistant.name, "audio": f},
        #     )
        #     if not response.is_success:
        #         raise APIError(f"Failed to teach the assistant: {response.text}")

    async def video(self, video: Union[str, Path]) -> None:
        """
        Learn the content of a video file.

        Args:
            video (Union[str, Path]): The path to the video file.
        """
        raise NotImplementedError()



import json
import re
from datetime import datetime
from typing import AsyncIterable, AsyncIterator, Iterable, Iterator, List, Optional

from firedust.types import (
    STRUCTURED_SCHEMA,
    AssistantConfig,
    Message,
    MessageStreamEvent,
    ReferencedMessage,
    StructuredAssistantMessage,
    StructuredUserMessage,
)
from firedust.types.chat import UserMessage
from firedust.utils.api import AsyncAPIClient, SyncAPIClient
from firedust.utils.errors import APIError


class Chat:
    """
    A collection of asynchronous methods to chat with the assistant.
    """

    def __init__(self, config: AssistantConfig, api_client: SyncAPIClient) -> None:
        self.config = config
        self.api_client = api_client
        self._previous_stream_chunk = ""

    def stream(
        self,
        message: str,
        chat_group: str = "default",
        username: Optional[str] = None,
        character: Optional[str] = None,
        add_to_memory: bool = True,
    ) -> Iterator[MessageStreamEvent]:
        """
        Streams the assistant's response to a message. The assistants uses relevant memories and
        previous conversations to generate a response. Conversations with the given user are private,
        and the assistant does not share the information with other users.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        query = "Summarize the research papers about..."

        for event in assistant.chat.stream(query):
            print(event.message)

        # See which memories the assistant used to generate the response. The references
        # are always provided in the last event of the stream.
        memory_ids = event.references.memories
        memories = assistant.memory.get(memory_ids)
        ```

        Args:
            message (str): The message to send.
            user (str, optional): The unique identifier of the user. Defaults to "default".

        Yields:
            MessageStreamEvent: The response from the assistant.
        """
        user_message = UserMessage(
            assistant=self.config.name,
            username=username,
            content=message,
            chat_group=chat_group,
            timestamp=datetime.now().timestamp(),
        )
        try:
            for msg in self.api_client.post_stream(
                "/assistant/chat/stream",
                data={
                    "message": user_message.model_dump(),
                    "response_config": {
                        "character": character,
                    },
                    "add_to_memory": add_to_memory,
                },
            ):
                msg_decoded = msg.decode("utf-8")
                for event in _process_stream_chunk(
                    msg_decoded, self._previous_stream_chunk
                ):
                    yield event
        except Exception as e:
            raise APIError(
                code=500,
                message=f"Failed to stream conversation: {e}",
            )
        finally:
            self._previous_stream_chunk = ""

    def message(
        self,
        message: str,
        chat_group: str = "default",
        username: Optional[str] = None,
        character: Optional[str] = None,
        add_to_memory: bool = True,
    ) -> ReferencedMessage:
        """
        Returns a full response from the assistant to a message. It takes longer to get a response
        compared to the stream method because the assistant processes the entire message before
        responding. The assistant uses relevant memories and previous conversations to generate a
        response. Conversations with the given user are private, and the assistant does not share the
        information with other users.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        query = "Summarize the research papers about..."

        response = assistant.chat.message(query)
        print(response.content)

        # See which memories the assistant used to generate the response.
        memory_ids = response.references.memories
        memories = assistant.memory.get(memory_ids)
        ```

        Args:
            content (str): The content of the message to send.
            chat_group (str, optional): The unique identifier of the chat group. Defaults to "default".
            username (str, optional): The username of the user. Defaults to None.
            character (str, optional): The name of the persona the assistant should embody in the response. Defaults to None.

        Returns:
            ReferencedMessage: The response from the assistant.
        """
        user_message = UserMessage(
            assistant=self.config.name,
            username=username,
            content=message,
            chat_group=chat_group,
            timestamp=datetime.now().timestamp(),
        )

        response = self.api_client.post(
            "/assistant/chat/message",
            data={
                "message": user_message.model_dump(),
                "response_config": {
                    "character": character,
                },
                "add_to_memory": add_to_memory,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to send the message: {response.text}",
            )
        return ReferencedMessage(**response.json()["data"])

    def structured(
        self,
        message: str,
        schema: STRUCTURED_SCHEMA,
        chat_group: str = "default",
        username: Optional[str] = None,
        character: Optional[str] = None,
        add_to_memory: bool = True,
    ) -> StructuredAssistantMessage:
        """
        Returns a json response that is structured based on the given schema.

        Example:
        ```python
        import firedust
        from firedust.types import ...

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        schema: STRUCTURED_SCHEMA = {
            "name": StringField(hint="The name of the person."),
            "occupation": StringField(hint="The occupation of the person."),
        }
        response = assistant.chat.structured(
            message="My name is John Doe and I'm a software engineer.",
            user="test_user",
            schema=schema,
        )
        print(response.message)
        ```

        Args:
            message (str): The message to send.
            schema (STRUCTURED_SCHEMA): The schema to structure the response.
            user (str, optional): The unique identifier of the user. Defaults to "default".
            add_to_memory (bool, optional): Whether to add the interaction to memory. Defaults to True.

        Returns:
            StructuredAssistantMessage: The structured response from the assistant.
        """
        user_message = StructuredUserMessage(
            assistant=self.config.name,
            chat_group=chat_group,
            username=username,
            content=message,
            schema_=schema,
            timestamp=datetime.now().timestamp(),
        )
        response = self.api_client.post(
            "/assistant/chat/structured",
            data={
                "add_to_memory": add_to_memory,
                "message": {**user_message.model_dump()},
                "response_config": {
                    "character": character,
                },
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to send the message: {response.text}",
            )

        return StructuredAssistantMessage(**response.json()["data"])

    def add_history(self, messages: Iterable[Message]) -> None:
        """
        Adds chat message history. It will become available for the assistant to
        learn from past conversations to improve the quality of responses.

        Example:
        ```python
        import firedust
        from firedust.types import Message

        assistant = firedust.assistant.load("ASSISTANT_NAME")

        message1 = Message(
            assistant="ASSISTANT_NAME",
            chat_group="product_team",
            message="John: Based on the last discussion, we've made the following changes to the product...",
            author="user",
        )
        message2 = Message(
            assistant="ASSISTANT_NAME",
            chat_group="product_team",
            message="Helen: John, could you please share the updated product roadmap?",
            author="user",
        )
        message3 = Message(
            assistant="ASSISTANT_NAME",
            chat_group="product_team",
            message="John: Sure, the new roadmap is the following...",
            author="user",
        )

        assistant.chat.add_history([message1, message2, message3])
        ```

        Args:
            messages (Iterable[Message]): The chat messages.
        """
        response = self.api_client.put(
            "/assistant/chat/history",
            data={
                "assistant": self.config.name,
                "messages": [msg.model_dump() for msg in messages],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to add chat history: {response.text}",
            )

    def erase_history(self, chat_group: str, confirm: bool = False) -> None:
        """
        Irreversebly delete the chat history of a chat group from the assistant's memory.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        assistant.chat.erase_history(chat_group="product_team", confirm=True)
        ```

        Args:
            chat_group (str): The unique identifier of the chat group.
            confirm (bool): Confirm the deletion. Defaults to False.
        """
        if confirm is False:
            raise ValueError("Please confirm the deletion by setting confirm=True")

        response = self.api_client.delete(
            "/assistant/chat/history",
            params={
                "assistant": self.config.name,
                "chat_group": chat_group,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to erase chat history: {response.text}",
            )

    def get_history(
        self, chat_group: str, limit: int = 25, offset: int = 0
    ) -> List[Message]:
        """
        Get the chat history of a user from the assistant's memory.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        history = assistant.chat.get_history(chat_group="product_team", limit=10)

        for message in history:
            print(message.message)
        ```

        Args:
            chat_group (str): The unique identifier of the chat group.
            limit (int): The maximum number of messages to return. Defaults to 25.
            offset (int): The number of messages to skip. Defaults to 0.

        Returns:
            List[Message]: The chat history of the user
        """
        response = self.api_client.get(
            "/assistant/chat/history",
            params={
                "assistant": self.config.name,
                "chat_group": chat_group,
                "limit": limit,
                "offset": offset,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to get chat history: {response.text}",
            )

        return [Message(**msg) for msg in response.json()["data"]]


class AsyncChat:
    """
    A collection of asynchronous methods to chat with the assistant.
    """

    def __init__(self, config: AssistantConfig, api_client: AsyncAPIClient) -> None:
        self.config = config
        self.api_client = api_client
        self._previous_stream_chunk = ""

    async def stream(
        self,
        message: str,
        chat_group: str = "default",
        username: Optional[str] = None,
        character: Optional[str] = None,
        add_to_memory: bool = True,
    ) -> AsyncIterator[MessageStreamEvent]:
        """
        Streams the assistant's response to a message. The assistants uses relevant memories and
        previous conversations to generate a response. Conversations with the given user are private and
        are not shared with other users.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
            query = "Summarize the research papers about..."

            async for event in assistant.chat.stream(query):
                print(event.message)

            # See which memories the assistant used to generate the response. The references
            # are always provided in the last event of the stream.
            memory_ids = event.references.memories
            memories = await assistant.memory.async_get(memory_ids)

        asyncio.run(main())
        ```

        Args:
            message (str): The message to send.
            user (str, optional): The unique identifier of the user. Defaults to "default".

        Yields:
            MessageStreamEvent: The response from the assistant.
        """
        user_message = UserMessage(
            assistant=self.config.name,
            chat_group=chat_group,
            username=username,
            content=message,
            timestamp=datetime.now().timestamp(),
        )
        try:
            async for msg in self.api_client.post_stream(
                "/assistant/chat/stream",
                data={
                    "message": user_message.model_dump(),
                    "response_config": {
                        "character": character,
                    },
                    "add_to_memory": add_to_memory,
                },
            ):
                msg_decoded = msg.decode("utf-8")
                async for event in _async_process_stream_chunk(
                    msg_decoded, self._previous_stream_chunk
                ):
                    yield event
        except Exception as e:
            raise APIError(
                code=500,
                message=f"Failed to stream the conversation: {e}",
            )
        finally:
            self._previous_stream_chunk = ""

    async def message(
        self,
        message: str,
        chat_group: str = "default",
        username: Optional[str] = None,
        character: Optional[str] = None,
        add_to_memory: bool = True,
    ) -> ReferencedMessage:
        """
        Returns a full response from the assistant to a message. It takes longer to get a response
        compared to the stream method because the assistant processes the entire message before
        responding. The assistant uses relevant memories and previous conversations to generate a
        response. Conversations with the given user are private and are not shared with other users.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
            query = "Summarize the research papers about..."

            response = await assistant.chat.message(query)
            print(response.message)

            # See which memories the assistant used to generate the response.
            memory_ids = response.references.memories
            memories = await assistant.memory.async_get(memory_ids)

        asyncio.run(main())
        ```

        Args:
            message (str): The message to send.
            chat_group (str, optional): The unique identifier of the chat group. Defaults to "default".
            username (str, optional): The unique identifier of the user. Defaults to "default".
            character (str, optional): The character of the assistant. Defaults to "default".
            add_to_memory (bool, optional): Whether to add the interaction to memory. Defaults to True.

        Returns:
            str: The response from the assistant.
        """
        user_message = UserMessage(
            assistant=self.config.name,
            chat_group=chat_group,
            username=username,
            content=message,
            timestamp=datetime.now().timestamp(),
        )
        response = await self.api_client.post(
            "/assistant/chat/message",
            data={
                "message": {**user_message.model_dump()},
                "response_config": {
                    "character": character,
                },
                "add_to_memory": add_to_memory,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to send the message: {response.text}",
            )

        return ReferencedMessage(**response.json()["data"])

    async def structured(
        self,
        message: str,
        schema: STRUCTURED_SCHEMA,
        chat_group: str = "default",
        username: Optional[str] = None,
        character: Optional[str] = None,
        add_to_memory: bool = True,
    ) -> StructuredAssistantMessage:
        """
        Returns a json response that is structured based on the given schema, asynchronously.

        Example:
        ```python
        import firedust
        from firedust.types import STRUCTURED_SCHEMA, StringField
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
            schema: STRUCTURED_SCHEMA = {
                "name": StringField(hint="The name of the person."),
                "occupation": StringField(hint="The occupation of the person."),
            }
            message = "My name is John Doe and I'm a software engineer."

            response = await assistant.chat.structured(message, schema)
            print(response.message)

        asyncio.run(main())
        ```

        Args:
            message (str): The message to send.
            schema (STRUCTURED_SCHEMA): The schema to structure the response.
            chat_group (str, optional): The unique identifier of the chat group. Defaults to "default".
            username (str, optional): The unique identifier of the user. Defaults to "default".
            character (str, optional): The character of the assistant. Defaults to "default".
            add_to_memory (bool, optional): Whether to add the interaction to memory. Defaults to True.

        Returns:
            StructuredAssistantMessage: The structured response from the assistant.
        """
        user_message = StructuredUserMessage(
            assistant=self.config.name,
            chat_group=chat_group,
            username=username,
            content=message,
            schema_=schema,
            timestamp=datetime.now().timestamp(),
        )
        response = await self.api_client.post(
            "/assistant/chat/structured",
            data={
                "add_to_memory": add_to_memory,
                "message": {**user_message.model_dump()},
                "response_config": {
                    "character": character,
                },
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to send the message: {response.text}",
            )

        return StructuredAssistantMessage(**response.json()["data"])

    async def add_history(self, messages: Iterable[Message]) -> None:
        """
        Adds a chat message history to the assistant's memory, asynchronously. It helps the assistant
        learn from past conversations to improve the quality of responses.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_NAME")

            message1 = Message(
                assistant="ASSISTANT_NAME",
                chat_group="product_team",
                message="John: Based on the last discussion, we've made the following changes to the product...",
                author="user",
            )
            message2 = Message(
                assistant="ASSISTANT_NAME",
                chat_group="product_team",
                message="Helen: John, could you please share the updated product roadmap?",
                author="user",
            )
            message3 = Message(
                assistant="ASSISTANT_NAME",
                chat_group="product_team",
                message="John: Sure, the new roadmap is the following...",
                author="user",
            )

            await assistant.chat.add_history([message1, message2, message3])

        asyncio.run(main())
        ```

        Args:
            messages (Iterable[Message]): The chat messages.
        """
        response = await self.api_client.put(
            "/assistant/chat/history",
            data={
                "assistant": self.config.name,
                "messages": [msg.model_dump() for msg in messages],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to add chat history: {response.text}",
            )

    async def erase_history(
        self,
        chat_group: str = "default",
        confirm: bool = False,
    ) -> None:
        """
        Irreversebly delete the chat history of a user from the assistant's memory, asynchronously.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_NAME")
            await assistant.chat.erase_history(chat_group="product_team", confirm=True)

        asyncio.run(main())
        ```

        Args:
            chat_group (str): The unique identifier of the chat group. Defaults to "default".
            confirm (bool): Confirm the deletion. Defaults to False.
        """
        if confirm is False:
            raise ValueError("Please confirm the deletion by setting confirm=True")

        response = await self.api_client.delete(
            "/assistant/chat/history",
            params={
                "assistant": self.config.name,
                "chat_group": chat_group,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to erase chat history: {response.text}",
            )

    async def get_history(
        self,
        chat_group: str = "default",
        limit: int = 25,
        offset: int = 0,
    ) -> List[Message]:
        """
        Get the chat history of a user from the assistant's memory, asynchronously.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_NAME")
            history = await assistant.chat.get_history(chat_group="product_team", limit=10)

            for message in history:
                print(message.message)

        asyncio.run(main())
        ```

        Args:
            chat_group (str): The unique identifier of the chat group. Defaults to "default".
            limit (int): The maximum number of messages to return. Defaults to 25.
            offset (int): The number of messages to skip. Defaults to 0.

        Returns:
            List[Message]: The chat history of the user
        """
        response = await self.api_client.get(
            "/assistant/chat/history",
            params={
                "assistant": self.config.name,
                "chat_group": chat_group,
                "limit": limit,
                "offset": offset,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to get chat history: {response.text}",
            )

        return [Message(**msg) for msg in response.json()["data"]]


def _process_stream_chunk(
    chunk: str, previous_chunk: str
) -> Iterable[MessageStreamEvent]:
    """
    Process a chunk of streamed data and yield MessageStreamEvent objects.

    Args:
        chunk (str): The current chunk of data.
        previous_chunk (str): The previous chunk of data, in case of incomplete data.

    Yields:
        MessageStreamEvent: The processed event from the data chunk.
    """
    chunk_split = re.split(r"\n\ndata: ", chunk)
    for data in chunk_split:
        if previous_chunk:
            data = previous_chunk + data
            previous_chunk = ""
        data = re.sub(r"^data: ", "", data).strip()
        try:
            yield MessageStreamEvent(**json.loads(data))
        except json.JSONDecodeError:
            previous_chunk = data


async def _async_process_stream_chunk(
    chunk: str, previous_chunk: str
) -> AsyncIterable[MessageStreamEvent]:
    """
    Asynchronously process a chunk of streamed data and yield MessageStreamEvent objects.

    Args:
        chunk (str): The current chunk of data.
        previous_chunk (str): The previous chunk of data, in case of incomplete data.

    Yields:
        MessageStreamEvent: The processed event from the data chunk.
    """
    chunk_split = re.split(r"\n\ndata: ", chunk)
    for data in chunk_split:
        if previous_chunk:
            data = previous_chunk + data
            previous_chunk = ""
        data = re.sub(r"^data: ", "", data).strip()
        try:
            yield MessageStreamEvent(**json.loads(data))
        except json.JSONDecodeError:
            previous_chunk = data



from typing import List
from uuid import UUID

import firedust
from firedust.types import APIContent, AssistantConfig, MemoryItem
from firedust.utils.api import AsyncAPIClient, SyncAPIClient
from firedust.utils.errors import APIError


class Memory:
    """
    A collection of methods to interact with the assistant's memory.
    """

    def __init__(self, config: AssistantConfig, api_client: SyncAPIClient) -> None:
        self.config = config
        self.api_client = api_client

    def recall(self, query: str, limit: int = 50, offset: int = 0) -> List[MemoryItem]:
        """
        Recall memories based on a query.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        memories = assistant.memory.recall("Information about late deliveries.")
        ```

        Args:
            query (str): The query to search memories for.
            limit (int): The maximum number of memories to return.
            offset (int): The offset to start from.

        Returns:
            List[MemoryItem]: The list of memories that are related to the query.
        """
        response = self.api_client.post(
            "/assistant/memory/recall",
            data={
                "assistant": self.config.name,
                "query": query,
                "limit": limit,
                "offset": offset,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to recall memories: {response.text}",
            )

        content = APIContent(**response.json())
        return [MemoryItem(**memory) for memory in content.data]

    def get(self, memory_ids: List[UUID]) -> List[MemoryItem]:
        """
        Retrieve a list of memory items by their IDs.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        response = assistant.chat.message("What are the latest sales figures?")

        # Retrieve memories that the assistant used to generate the response
        memory_ids = response.references.memories
        memories = assistant.memory.get(memory_ids)
        ```

        Args:
            memory_ids (List[UUID]): A list of memory IDs.

        Returns:
            List[MemoryItem]: A list of memory items.
        """
        if len(memory_ids) == 0:
            return []

        response = self.api_client.post(
            "/assistant/memory/list",
            data={
                "assistant": self.config.name,
                "memory_ids": [str(memory_id) for memory_id in memory_ids],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to get memories: {response.text}",
            )

        content = APIContent(**response.json())
        return [MemoryItem(**memory) for memory in content.data]

    def add(self, memories: List[MemoryItem]) -> None:
        """
        Add new memory items to the assistant's memory collection. It is useful for memory management,
        for example, when selectively migrating some memories from one assistant to another.

        Example:
        ```python
        import firedust

        assistant1 = firedust.assistant.load("ASSISTANT_NAME1")
        assistant2 = firedust.assistant.load("ASSISTANT_NAME2")

        memories = assistant1.memory.recall("User feedback")
        assistant2.memory.add(memories)
        ```

        Args:
            memories (List[MemoryItem]): The list of memory items to add.
        """
        response = self.api_client.put(
            "/assistant/memory/list",
            data={
                "assistant": self.config.name,
                "memories": [memory.model_dump() for memory in memories],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to add memories: {response.text}",
            )

    def delete(self, memory_ids: List[UUID]) -> None:
        """
        Removes memories from the assistant's memory. It is useful for memory management,
        for example, when selectively removing memories from an assistant.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        memory_ids = assistant.memory.recall("User feedback")
        assistant.memory.delete(memory_ids)
        ```

        Args:
            memory_ids (List[UUID]): The list of memory IDs to remove.
        """
        response = self.api_client.post(
            "/assistant/memory/delete",
            data={
                "assistant": self.config.name,
                "memory_ids": [str(memory_id) for memory_id in memory_ids],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to remove memory: {response.text}",
            )

    def list(self, limit: int = 100, offset: int = 0) -> List[UUID]:
        """
        List all memory items available to the assistant.

        Returns:
            List[UUID]: A list of memory IDs.
        """
        response = self.api_client.get(
            "/assistant/memory/list",
            params={"assistant": self.config.name, "limit": limit, "offset": offset},
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to list memories: {response.text}",
            )

        content = APIContent(**response.json())
        return [UUID(memory_id) for memory_id in content.data]

    def share(self, assistant_receiver: str) -> None:
        """
        Share all assistant's memories with another assistant. It makes the memories
        available for the other assistant to use when answering questions or performing tasks.
        It's a full sync and applies to both past memories and the ones added in the future.

        To share memories selectively use the assistant.memory.add method.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_SHARER")
        assistant.memory.share("ASSISTANT_RECEIVER")
        ```

        Args:
            assistant_receiver (str): The name of the assistant who will access the memories.
        """
        if assistant_receiver == self.config.name:
            raise ValueError("Cannot attach memories from the same assistant.")

        receiver_config = firedust.assistant.load(assistant_receiver)
        if self.config.name in receiver_config.config.attached_memories:
            raise ValueError(
                f"The memories of {self.config.name} are already shared to the {assistant_receiver}."
            )

        response = self.api_client.put(
            "/assistant/memory/share",
            data={
                "assistant_sharer": self.config.name,
                "assistant_receiver": assistant_receiver,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to share memories: {response.text}",
            )

    def unshare(self, assistant_receiver: str) -> None:
        """
        Memories of the assistant will no longer be shared with the assistant receiver.

        Example:
        ```python
        import firedust

        assistant_sharer = firedust.assistant.load("ASSISTANT_SHARER")
        assistant_sharer.memory.unshare("ASSISTANT_RECEIVER")
        ```

        Args:
            assistant (str): The name of the assistant who will lose access to the memories.
        """
        if assistant_receiver == self.config.name:
            raise ValueError("Cannot detach memories from the same assistant.")

        reciver_config = firedust.assistant.load(assistant_receiver)
        if self.config.name not in reciver_config.config.attached_memories:
            raise ValueError(
                f"The memories of {self.config.name} are not shared to the {assistant_receiver}."
            )

        response = self.api_client.delete(
            "/assistant/memory/share",
            params={
                "assistant_receiver": assistant_receiver,
                "assistant_sharer": self.config.name,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to detach collection: {response.text}",
            )


class AsyncMemory:
    """
    A collection of asynchronous methods to interact with the assistant's memory asynchronously.
    """

    def __init__(self, config: AssistantConfig, api_client: AsyncAPIClient) -> None:
        self.config = config
        self.api_client = api_client

    async def recall(
        self, query: str, limit: int = 50, offset: int = 0
    ) -> List[MemoryItem]:
        """
        Recall memories based on a query, asynchronously.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_NAME")
            memories = await assistant.memory.recall("Information about late deliveries.")

        asyncio.run(main())
        ```

        Args:
            query (str): The query to search memories for.
            limit (int): The maximum number of memories to return.
            offset (int): The offset to start from.

        Returns:
            List[MemoryItem]: The list of memories that are related to the query.
        """
        response = await self.api_client.post(
            "/assistant/memory/recall",
            data={
                "assistant": self.config.name,
                "query": query,
                "limit": limit,
                "offset": offset,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to recall memories: {response.text}",
            )

        content = APIContent(**response.json())
        return [MemoryItem(**memory) for memory in content.data]

    async def get(self, memory_ids: List[UUID]) -> List[MemoryItem]:
        """
        Retrieve a list of memory items by their IDs, asynchronously. It is used for memory management.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_NAME")
            response = await assistant.chat.message("What are the latest sales figures?")

            # Retrieve memories that the assistant used to answer the question
            memory_ids = response.references.memories
            memories = await assistant.memory.get(memory_ids)

        asyncio.run(main())
        ```

        Args:
            memory_ids (List[UUID]): A list of memory IDs.

        Returns:
            List[MemoryItem]: A list of memory items.
        """
        response = await self.api_client.post(
            "/assistant/memory/list",
            data={
                "assistant": self.config.name,
                "memory_ids": [str(memory_id) for memory_id in memory_ids],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to get memories: {response.text}",
            )

        content = APIContent(**response.json())
        return [MemoryItem(**memory) for memory in content.data]

    async def add(self, memories: List[MemoryItem]) -> None:
        """
        Adds new memory items to the assistant's memory collection, asynchronously. It is useful for memory management.
        For example, when selectively migrating some memories from one assistant to another.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant1 = await firedust.async_load("ASSISTANT_NAME1")
            assistant2 = await firedust.async_load("ASSISTANT_NAME2")

            memories = await assistant1.memory.recall("User feedback")
            await assistant2.memory.add(memories)

        asyncio.run(main())
        ```

        Args:
            memories (List[MemoryItem]): The list of memory items to add.
        """
        response = await self.api_client.put(
            "/assistant/memory/list",
            data={
                "assistant": self.config.name,
                "memories": [memory.model_dump() for memory in memories],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to add memories: {response.text}",
            )

    async def delete(self, memory_ids: List[UUID]) -> None:
        """
        Removes memories from the assistant's default memory collection, async. It is useful for memory management,
        for example, when selectively removing memories from an assistant.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_NAME")
            memory_ids = await assistant.memory.list(limit=10)
            await assistant.memory.delete(memory_ids)

        asyncio.run(main())
        ```

        Args:
            memory_ids (List[UUID]): The list of memory IDs to remove.
        """
        response = await self.api_client.post(
            "/assistant/memory/delete",
            data={
                "assistant": self.config.name,
                "memory_ids": [str(memory_id) for memory_id in memory_ids],
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to delete memories: {response.text}",
            )

    async def list(self, limit: int = 100, offset: int = 0) -> List[UUID]:
        """
        List all memory items available to the assistant.

        Returns:
            List[UUID]: A list of memory IDs.
        """
        response = await self.api_client.get(
            "/assistant/memory/list",
            params={"assistant": self.config.name, "limit": limit, "offset": offset},
        )

        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to list memories: {response.text}",
            )

        content = APIContent(**response.json())
        return [UUID(memory_id) for memory_id in content.data]

    async def share(self, assistant_receiver: str) -> None:
        """
        Share all assistant's memories with another assistant, asynchronously. It makes the memories
        available for the assistant receiver to use when answering questions or performing tasks.
        It's a full memory sync and applies to both past memories and the ones added in the future.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_SHARER")
            await assistant.memory.share("ASSISTANT_RECEIVER")

        asyncio.run(main())
        ```

        Args:
            assistant_receiver (str): The name of the assistant who will access the memories.
        """
        if assistant_receiver == self.config.name:
            raise ValueError("Cannot share memories from the same assistant.")

        receiver_config = await firedust.assistant.async_load(assistant_receiver)
        if self.config.name in receiver_config.config.attached_memories:
            raise ValueError(
                f"The memories of {self.config.name} are already shared to the {assistant_receiver}."
            )

        response = await self.api_client.put(
            "/assistant/memory/share",
            data={
                "assistant_receiver": assistant_receiver,
                "assistant_sharer": self.config.name,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to share memories: {response.text}",
            )

    async def unshare(self, assistant_receiver: str) -> None:
        """
        Memories of the assistant will no longer be shared with the assistant receiver.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.async_load("ASSISTANT_SHARER")
            await assistant.memory.unshare("ASSISTANT_RECEIVER")

        asyncio.run(main())
        ```

        Args:
            assistant (str): The name of the assistant who will lose access to the memories.
        """
        if assistant_receiver == self.config.name:
            raise ValueError("Cannot unshare memories from the same assistant.")

        reciver_config = await firedust.assistant.async_load(assistant_receiver)
        if self.config.name not in reciver_config.config.attached_memories:
            raise ValueError(
                f"The memories of {self.config.name} are not shared with {assistant_receiver}."
            )

        response = await self.api_client.delete(
            "/assistant/memory/share",
            params={
                "assistant_sharer": self.config.name,
                "assistant_receiver": assistant_receiver,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to unshare collection: {response.text}",
            )



import httpx

from firedust.types import AssistantConfig
from firedust.types.interface import SlackConfig, SlackTokens
from firedust.utils.api import AsyncAPIClient, SyncAPIClient
from firedust.utils.errors import SlackError


class SlackInterface:
    """
    Slack interface module for the Firedust assistant. Deploy the assistant to Slack
    to interact with it in your workspace.

    Example:
    https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack.py
    """

    def __init__(self, config: AssistantConfig, api_client: SyncAPIClient) -> None:
        self.config: AssistantConfig = config
        self.api_client: SyncAPIClient = api_client

    def create_app(
        self, description: str, greeting: str, configuration_token: str
    ) -> httpx.Response:
        """
        Creates a Slack app for the assistant.

        Example:
        https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack.py

        Args:
            description (str): A short description of the assistant.
            greeting (str): A message that the assistant will greet users with.
            configuration_token (str): The Slack configuration token.

        Returns:
            httpx.Response: The response from the API.
        """
        response = self.api_client.post(
            "/assistant/interface/slack",
            data={
                "assistant": self.config.name,
                "description": description,
                "greeting": greeting,
                "configuration_token": configuration_token,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to create Slack app: {response.text}")

        self.config.interfaces.slack = SlackConfig(**response.json()["data"])

        return response

    def set_tokens(self, app_token: str, bot_token: str) -> httpx.Response:
        """
        Sets the Slack tokens.

        Example:
        https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack.py

        Args:
            app_token (str): The Slack app token.
            bot_token (str): The Slack bot token.

        Returns:
            httpx.Response: The response from the API.
        """
        if self.config.interfaces.slack is None:
            raise SlackError(
                "Slack configuration not found. Please, create an app first. Hint: assistant.interfece.slack.create_app()"
            )

        response = self.api_client.put(
            "/assistant/interface/slack/tokens",
            data={
                "assistant": self.config.name,
                "app_token": app_token,
                "bot_token": bot_token,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to set Slack tokens: {response.text}")

        self.config.interfaces.slack.tokens = SlackTokens(
            app_token=app_token, bot_token=bot_token
        )
        return response

    def set_greeting(self, greeting: str) -> httpx.Response:
        """
        Sets the Slack greeting message.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        assistant.interface.slack.set_greeting("Hello! I'm here to help you.")
        ```

        Args:
            greeting (str): The greeting message.
        """
        if self.config.interfaces.slack is None:
            raise SlackError(
                "Slack configuration not found. Please, create an app first. Hint: assistant.interface.slack.create_app()"
            )

        response = self.api_client.put(
            "/assistant/interface/slack/greeting",
            data={
                "assistant": self.config.name,
                "greeting": greeting,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to set Slack greeting: {response.text}")

        self.config.interfaces.slack.greeting = greeting
        return response

    def deploy(self) -> httpx.Response:
        """
        Deploys the assistant to Slack.

        Example:
        https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack.py
        """
        response = self.api_client.put(
            "/assistant/interface/slack/deploy",
            data={
                "assistant": self.config.name,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to deploy Slack app: {response.text}")

        return response

    def remove_deployment(self) -> httpx.Response:
        """
        Removes the assistant deployment from Slack.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        assistant.interface.slack.remove_deployment()
        ```
        """
        response = self.api_client.delete(
            "/assistant/interface/slack/deploy",
            params={"assistant": self.config.name},
        )
        if not response.is_success:
            raise SlackError(f"Failed to remove deployment from Slack: {response.text}")

        return response

    def delete_app(self, configuration_token: str) -> httpx.Response:
        """
        Deletes the Slack app. Requires a configuration token from your Slack workspace.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        assistant.interface.slack.delete_app("SLACK_CONFIGURATION_TOKEN")
        ```

        Args:
            configuration_token (str): The Slack configuration token.
        """
        response = self.api_client.delete(
            "/assistant/interface/slack",
            params={
                "assistant": self.config.name,
                "configuration_token": configuration_token,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to delete Slack interface: {response.text}")

        return response


class AsyncSlackInterface:
    """
    Async Slack interface module for the Firedust assistant.

    Example:
    https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack_async.py
    """

    def __init__(self, config: AssistantConfig, api_client: AsyncAPIClient) -> None:
        self.config: AssistantConfig = config
        self.api_client: AsyncAPIClient = api_client

    async def create_app(
        self, description: str, configuration_token: str
    ) -> httpx.Response:
        """
        Creates a Slack app for the assistant asynchronously.

        Example:
        https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack_async.py

        Args:
            description (str): A short description of the assistant.
            configuration_token (str): The Slack configuration token.
        """
        response = await self.api_client.post(
            "/assistant/interface/slack",
            data={
                "assistant": self.config.name,
                "description": description,
                "configuration_token": configuration_token,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to create Slack app: {response.text}")

        self.config.interfaces.slack = SlackConfig(**response.json()["data"])

        return response

    async def set_tokens(self, app_token: str, bot_token: str) -> httpx.Response:
        """
        Sets the Slack tokens.

        Example:
        https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack_async.py

        Args:
            app_token (str): The Slack app token.
            bot_token (str): The Slack bot token.
        """
        if self.config.interfaces.slack is None:
            raise SlackError(
                "Slack configuration not found. Please, create an app first. Hint: assistant.interface.slack.create_app()"
            )

        response = await self.api_client.put(
            "/assistant/interface/slack/tokens",
            data={
                "assistant": self.config.name,
                "app_token": app_token,
                "bot_token": bot_token,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to set Slack tokens: {response.text}")

        self.config.interfaces.slack.tokens = SlackTokens(
            app_token=app_token, bot_token=bot_token
        )
        return response

    async def set_greeting(self, greeting: str) -> httpx.Response:
        """
        Sets the Slack greeting message.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.load("ASSISTANT_NAME")
            await assistant.interface.slack.set_greeting("Hello! I'm here to help you.")

        asyncio.run(main())
        ```

        Args:
            greeting (str): The greeting message.
        """
        if self.config.interfaces.slack is None:
            raise SlackError(
                "Slack configuration not found. Please, create an app first. Hint: assistant.interface.slack.create_app()"
            )

        response = await self.api_client.put(
            "/assistant/interface/slack/greeting",
            data={
                "assistant": self.config.name,
                "greeting": greeting,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to set Slack greeting: {response.text}")

        self.config.interfaces.slack.greeting = greeting
        return response

    async def deploy(self) -> httpx.Response:
        """
        Deploys the assistant to Slack.

        Example:
        https://github.com/ion2088/firedust/blob/master/examples/deploy_to_slack_async.py
        """
        response = await self.api_client.put(
            f"/assistant/interface/slack/deploy?assistant={self.config.name}"
        )
        if not response.is_success:
            raise SlackError(f"Failed to deploy Slack app: {response.text}")

        return response

    async def remove_deployment(self) -> httpx.Response:
        """
        Removes the assistant deployment from Slack.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.load("ASSISTANT_NAME")
            await assistant.interface.slack.remove_deployment()

        asyncio.run(main())
        """
        response = await self.api_client.delete(
            "/assistant/interface/slack/deploy",
            params={"assistant": self.config.name},
        )
        if not response.is_success:
            raise SlackError(f"Failed to remove deployment from Slack: {response.text}")

        return response

    async def delete_app(self, configuration_token: str) -> httpx.Response:
        """
        Deletes the Slack app. Requires a configuration token from your Slack workspace.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.load("ASSISTANT_NAME")
            await assistant.interface.slack.delete_app("SLACK_CONFIGURATION_TOKEN")

        asyncio.run(main())
        """
        response = await self.api_client.delete(
            "/assistant/interface/slack",
            params={
                "assistant": self.config.name,
                "configuration_token": configuration_token,
            },
        )
        if not response.is_success:
            raise SlackError(f"Failed to delete Slack interface: {response.text}")

        return response



from typing import Any, Literal
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_serializer

UNIX_TIMESTAMP = float  # see: https://www.unixtimestamp.com/
STREAM_STOP_EVENT = "[[STOP]]"
TOKENS = int
INFERENCE_MODEL = Literal[
    "mistral/mistral-medium",
    "mistral/mistral-small",
    "mistral/mistral-tiny",
    "openai/gpt-4",
    "openai/gpt-4-turbo-preview",
    "openai/gpt-4o",
    "openai/gpt-4o-mini",
    "openai/gpt-3.5-turbo",
]


class BaseConfig(BaseModel):
    """
    Represents the base configuration model.
    All configuration models should inherit from this class.
    """

    id: UUID = Field(default_factory=uuid4)

    def __setattr__(self, key: str, value: Any) -> None:
        # set immutable attributes
        if key == "id":
            raise AttributeError("Cannot set attribute 'id', it is immutable.")

        return super().__setattr__(key, value)

    @field_serializer("id", when_used="always")
    def serialize_config_id(self, value: UUID) -> str:
        return str(value)



from typing import Union


def is_unix_timestamp(timestamp: Union[int, float]) -> bool:
    """
    Check if the provided integer is a valid UNIX timestamp.

    UNIX timestamps are typically in the range from 0 to 2**31 - 1,
    covering dates from 1970-01-01 to 2038-01-19.

    Source: https://www.unixtimestamp.com/

    Args:
        timestamp (int): The integer to check.

    Returns:
        bool: True if the integer is a valid UNIX timestamp, False otherwise.
    """

    MIN_UNIX_TIMESTAMP = 0  # 1 January 1970
    MAX_UNIX_TIMESTAMP = 2**31 - 1  # 19 January 2038 for a 32-bit signed integer
    return MIN_UNIX_TIMESTAMP <= timestamp <= MAX_UNIX_TIMESTAMP



from typing import Any, Literal, Optional

from pydantic import BaseModel

from .base import UNIX_TIMESTAMP


class APIContent(BaseModel):
    """
    Represents the API content model.
    """

    status: Literal["success", "error"]
    timestamp: UNIX_TIMESTAMP
    data: Any = {}
    message: Optional[str] = None



# isort: skip_file
"""
Make the main types easily accessible.

```python
from firedust.types import Assistant, AssistantConfig, Message, ReferencedMessage, MessageStreamEvent, StructuredAssistantMessage, StructuredUserMessage, STRUCTURED_SCHEMA, BooleanField, DictField, FloatField, ListField, CategoryField, StringField, MemoryItem, APIContent, INFERENCE_MODEL

def custom_funk(assistant: Assistant) -> ...:
    ...
```
"""

from firedust.types.assistant import AssistantConfig as AssistantConfig
from firedust.types.chat import (
    Message as Message,
    ReferencedMessage as ReferencedMessage,
    MessageStreamEvent as MessageStreamEvent,
    StructuredAssistantMessage as StructuredAssistantMessage,
    StructuredUserMessage as StructuredUserMessage,
)
from firedust.types.structured import (
    STRUCTURED_SCHEMA as STRUCTURED_SCHEMA,
    BooleanField as BooleanField,
    DictField as DictField,
    FloatField as FloatField,
    ListField as ListField,
    CategoryField as CategoryField,
    StringField as StringField,
)
from firedust.types.base import INFERENCE_MODEL as INFERENCE_MODEL
from firedust.types.memory import MemoryItem as MemoryItem
from firedust.types.api import APIContent as APIContent
from firedust._assistant.base.assistant import (
    Assistant as Assistant,
    AsyncAssistant as AsyncAssistant,
)

__all__ = [
    "Assistant",
    "AsyncAssistant",
    "AssistantConfig",
    "Message",
    "ReferencedMessage",
    "MessageStreamEvent",
    "MemoryItem",
    "APIContent",
    "STRUCTURED_SCHEMA",
    "BooleanField",
    "DictField",
    "FloatField",
    "ListField",
    "CategoryField",
    "StringField",
    "StructuredAssistantMessage",
    "StructuredUserMessage",
    "INFERENCE_MODEL",
]



"""
The main entry point to create, load, and list your AI assistants.

Example:
    ```python
    import firedust

    # create a new assistant and interact with it
    assistant = firedust.assistant.create(
        name="Sam",
        instructions="1. Protect the ring bearer. 2. Do not let the ring corrupt you.",
    )
    response = assistant.chat.message("What is the shortest way to Mordor?")
    print(response.message)

    # stream messages
    for event in assistant.chat.stream("What is the shortest way to Mordor?"):
        print(event.message)

    # add your data to the assistant's memory
    important_data = [
        "Do not trust Smeagol.",
        "The ring is evil, it must be destroyed.",
    ]
    for data in important_data:
        assistant.learn.fast(data)
    response = assistant.chat.message("What should I do with the ring?")
    assert "destroy" in response.message.lower()

    # list all assistants
    assistants = firedust.assistant.list()

    # load an existing assistant
    assistant = firedust.assistant.load("Sam")
    ```

    For more examples, check:
    https://github.com/ion2088/firedust/tree/master/examples
"""

from typing import List

from firedust.types import APIContent, Assistant, AssistantConfig, AsyncAssistant
from firedust.types.base import INFERENCE_MODEL
from firedust.utils.api import AsyncAPIClient, SyncAPIClient
from firedust.utils.errors import APIError
from firedust.utils.logging import LOG


def create(
    name: str,
    instructions: str = "",
    model: INFERENCE_MODEL = "openai/gpt-4o",
) -> Assistant:
    """
    Creates a new assistant with the specified configuration.

    Args:
        name (str): The name of the assistant.
        instructions (str): The instructions for the assistant.
        model (INFERENCE_MODEL, optional): The inference model to use. Defaults to "openai/gpt-4".

    Returns:
        Assistant: A new instance of the assistant class.
    """
    config = AssistantConfig(name=name, instructions=instructions, model=model)
    api_client = SyncAPIClient()

    response = api_client.post("/assistant", data=config.model_dump())
    if not response.is_success:
        api_client.close()
        raise APIError(
            code=response.status_code,
            message=f"Failed to create an assistant with config {config}: {response.text}",
        )
    LOG.info(
        f"Assistant {config.name} was created successfully and saved to the cloud."
    )
    return Assistant._create_instance(config, api_client)


async def async_create(
    name: str,
    instructions: str = "",
    model: INFERENCE_MODEL = "openai/gpt-4o",
) -> AsyncAssistant:
    """
    Asynchronously creates a new assistant with the specified configuration.

    Args:
        name (str): The name of the assistant.
        instructions (str): The instructions for the assistant.
        model (INFERENCE_MODEL, optional): The inference model to use. Defaults to "openai/gpt-4".

    Returns:
        AsyncAssistant: A new instance of the AsyncAssistant class.
    """
    config = AssistantConfig(name=name, instructions=instructions, model=model)
    api_client = AsyncAPIClient()

    response = await api_client.post("/assistant", data=config.model_dump())
    if not response.is_success:
        await api_client.close()
        raise APIError(
            code=response.status_code,
            message=f"Failed to create an assistant with config {config}: {response.text}",
        )
    LOG.info(
        f"Assistant {config.name} was created successfully and saved to the cloud."
    )
    return await AsyncAssistant._create_instance(config, api_client)


def load(name: str) -> Assistant:
    """
    Loads an existing assistant with the specified name.

    Args:
        name (str): The name of the assistant to load.

    Returns:
        Assistant: A new instance of the Assistant class.
    """
    api_client = SyncAPIClient()
    response = api_client.get("/assistant", params={"name": name})
    if not response.is_success:
        api_client.close()
        raise APIError(
            code=response.status_code,
            message=f"Failed to load assistant {name}: {response.text}",
        )
    content = APIContent(**response.json())
    config = AssistantConfig(**content.data)
    return Assistant._create_instance(config, api_client)


async def async_load(name: str) -> AsyncAssistant:
    """
    Asynchronously loads an existing assistant with the specified name.

    Args:
        name (str): The name of the assistant to load.

    Returns:
        AsyncAssistant: A new instance of the AsyncAssistant class.
    """
    api_client = AsyncAPIClient()
    response = await api_client.get("/assistant", params={"name": name})
    if not response.is_success:
        await api_client.close()
        raise APIError(
            code=response.status_code,
            message=f"Failed to load the assistant with id {name}: {response.text}",
        )
    content = APIContent(**response.json())
    config = AssistantConfig(**content.data)
    return await AsyncAssistant._create_instance(config, api_client)


def list() -> List[Assistant]:
    """
    Lists all existing assistants.

    Returns:
        List[Assistant]: A list of Assistant objects.
    """
    api_client = SyncAPIClient()
    response = api_client.get("/assistant/list")
    if not response.is_success:
        api_client.close()
        raise APIError(
            code=response.status_code,
            message=f"Failed to list the assistants: {response.text}",
        )
    content = APIContent(**response.json())
    configs = [AssistantConfig(**assistant) for assistant in content.data]
    return [Assistant._create_instance(config, api_client) for config in configs]


async def async_list() -> List[AsyncAssistant]:
    """
    Asynchronously lists all existing assistants.

    Returns:
        List[AsyncAssistant]: A list of AsyncAssistant objects.
    """
    api_client = AsyncAPIClient()
    response = await api_client.get("/assistant/list")
    if not response.is_success:
        await api_client.close()
        raise APIError(
            code=response.status_code,
            message=f"Failed to list the assistants: {response.text}",
        )
    content = APIContent(**response.json())
    configs = [AssistantConfig(**assistant) for assistant in content.data]
    return [
        await AsyncAssistant._create_instance(config, api_client) for config in configs
    ]



from typing import Any, Dict, List, Literal, Mapping, Optional, Union

from pydantic import BaseModel


class BooleanResponse(BaseModel):
    response: bool
    conviction: float


class BaseField(BaseModel):
    type: Literal["float", "bool", "str", "list", "dict", "category"]
    hint: str
    optional: bool = False


class FloatField(BaseField):
    type: Literal["float"] = "float"
    min_: Optional[float] = None
    max_: Optional[float] = None


class BooleanField(BaseField):
    type: Literal["bool"] = "bool"


class StringField(BaseField):
    type: Literal["str"] = "str"


class CategoryField(BaseField):
    type: Literal["category"] = "category"
    categories: List[str]


class ListField(BaseField):
    type: Literal["list"] = "list"
    items: Union[FloatField, BooleanField, StringField, "ListField", "DictField"]


class DictField(BaseField):
    type: Literal["dict"] = "dict"
    items: Dict[
        str,
        Union[
            FloatField, BooleanField, StringField, CategoryField, ListField, "DictField"
        ],
    ]


FIELD_TYPE = Union[
    FloatField, BooleanField, StringField, CategoryField, ListField, DictField
]
ALLOWED_PRIMATY_VALUE_TYPES = Union[float, bool, str, None]
ALLOWED_VALUE_TYPES = Union[
    float,
    bool,
    str,
    List[Any],  # List[ALLOWED_VALUE_TYPES],
    Dict[str, Any],  # Dict[str, ALLOWED_VALUE_TYPES],
    None,
]
STRUCTURED_RESPONSE = Mapping[str, ALLOWED_VALUE_TYPES]
STRUCTURED_SCHEMA = Mapping[str, FIELD_TYPE]



from datetime import datetime
from typing import List, Literal, Optional, Sequence
from uuid import UUID

from pydantic import BaseModel, Field, field_serializer

from .base import UNIX_TIMESTAMP, BaseConfig
from .structured import STRUCTURED_RESPONSE, STRUCTURED_SCHEMA


class Message(BaseConfig):
    """
    Represents a message between a user and an assistant.
    """

    assistant: str = Field(..., description="The name of the assistant.")
    chat_group: str = Field(
        "default",
        description="The unique identifier of the chat group. All messages in a group are private.",
    )
    name: Optional[str] = Field(
        None,
        description="A name to diferentiate between users within the same chat_group. Applicable to both users and assistants.",
    )
    timestamp: UNIX_TIMESTAMP = Field(
        default_factory=lambda: datetime.now().timestamp(),
        description="The timestamp of the message.",
    )
    content: str = Field(..., description="The text of the message.")
    author: Literal["user", "assistant"] = Field(
        ..., description="The author of the message."
    )


class UserMessage(Message):
    author: Literal["user"] = Field(
        "user",
        description="Indicates that the author of the message is the user.",
    )


class StructuredUserMessage(UserMessage):
    schema_: STRUCTURED_SCHEMA = Field(
        ..., description="The schema of the structured request."
    )


class AssistantMessage(Message):
    author: Literal["assistant"] = Field(
        "assistant",
        description="Indicates that the author of the message is the assistant.",
    )


class MessageReferences(BaseModel):
    """
    Represents the references to the memories and messages used by the assistant to generate a response.
    """

    memory_ids: Sequence[UUID] = Field(
        ..., description="The references to the relevant memories."
    )
    message_ids: Sequence[UUID] = Field(
        ..., description="The references to the relevant conversations."
    )

    @field_serializer("memory_ids", when_used="always")
    def serialize_memory_refs(self, value: Sequence[UUID]) -> Sequence[str]:
        return [str(ref) for ref in value]

    @field_serializer("message_ids", when_used="always")
    def serialize_conversation_refs(self, value: Sequence[UUID]) -> Sequence[str]:
        return [str(ref) for ref in value]


class ReferencedMessage(Message):
    references: Optional[MessageReferences] = Field(
        None,
        description="References to memories and conversations used by the assistant as context to answer the query.",
    )


class StructuredAssistantMessage(BaseConfig):
    # Unable to build of AssistantMessage because we redefine message type
    assistant: str = Field(..., description="The name of the assistant.")
    chat_group: str = Field(
        "default",
        description="The unique identifier of the chat group. All messages in a group are private.",
    )
    name: Optional[str] = Field(
        None,
        description="A username to diferentiate between users with the same chat_group. Applicable to both users and assistants.",
    )
    timestamp: UNIX_TIMESTAMP = Field(
        default_factory=lambda: datetime.now().timestamp(),
        description="The timestamp of the message.",
    )
    content: STRUCTURED_RESPONSE = Field(
        ..., description="The schema of the structured request."
    )
    author: Literal["assistant"] = Field(
        "assistant",
        description="Indicates that the author of the message is the assistant.",
    )
    references: Optional[MessageReferences] = Field(
        None,
        description="References to memories and conversations used by the assistant as context to answer the query.",
    )


class MessageStreamEvent(ReferencedMessage):
    stream_ended: bool = Field(
        ..., description="Indicates whether the message stream has ended."
    )


class MessageContext(MessageReferences):
    """
    Represents the context of a message.
    """

    message_id: UUID = Field(..., description="The unique identifier of the message.")
    messages: List[Message] = Field(..., description="The messages in the chat.")
    instructions: str = Field(
        ..., description="The instructions related to the message."
    )

    @field_serializer("message_id", when_used="always")
    def serialize_message_id(self, value: UUID) -> str:
        return str(value)


# from datetime import datetime
# from typing import Literal, Optional, Sequence, Union
# from uuid import UUID

# from pydantic import BaseModel, Field, field_serializer

# from .base import UNIX_TIMESTAMP, BaseConfig
# from .structured import STRUCTURED_RESPONSE, STRUCTURED_SCHEMA


# class Message(BaseConfig, frozen=True):
#     """
#     Represents a message between a user and an assistant.

#     Args:
#         assistant (str): The name of the assistant.
#         user (str): The unique identifier of the user.
#         timestamp (UNIX_TIMESTAMP): The timestamp of the message.
#         message (str): The text of the message.
#         author (Literal["user", "assistant"]): The author of the message.
#     """

#     assistant: str = Field(...)
#     user: str = Field(...)
#     timestamp: UNIX_TIMESTAMP = Field(
#         default_factory=lambda: datetime.now().timestamp()
#     )
#     message: str = Field(...)
#     author: Literal["user", "assistant"] = Field(...)


# class UserMessage(Message):
#     author: Literal["user"] = Field(default="user")


# class StructuredUserMessage(UserMessage):
#     schema_: STRUCTURED_SCHEMA = Field(...)


# class AssistantMessage(Message):
#     author: Literal["assistant"] = Field(default="assistant")


# class MessageReferences(BaseModel):
#     """
#     References to the memories and messages used by the assistant as context to generate a response.

#     Args:
#         memories (Sequence[UUID]): The references to the relevant memories.
#         conversations (Sequence[UUID]): The references to the relevant conversations.
#     """

#     memories: Sequence[UUID] = Field(...)
#     conversations: Sequence[UUID] = Field(...)

#     @field_serializer("memories", when_used="always")
#     def serialize_memory_refs(self, value: Sequence[UUID]) -> Sequence[str]:
#         return [str(ref) for ref in value]

#     @field_serializer("conversations", when_used="always")
#     def serialize_conversation_refs(self, value: Sequence[UUID]) -> Sequence[str]:
#         return [str(ref) for ref in value]


# class StructuredAssistantMessage(BaseConfig, frozen=True):
#     assistant: str = Field(...)
#     user: str = Field(...)
#     timestamp: UNIX_TIMESTAMP = Field(...)
#     message: STRUCTURED_RESPONSE = Field(...)
#     author: Literal["assistant"] = Field(default="assistant")
#     references: Union[MessageReferences, None] = Field(default=None)


# class ReferencedMessage(Message, frozen=True):
#     references: Optional[MessageReferences] = Field(default=None)


# class MessageStreamEvent(ReferencedMessage, frozen=True):
#     stream_ended: bool = Field(...)



"""
This is a private file. To interact with Assistant and AsyncAssistant classes, use the modules in firedust.assistant.*.

Example:
```python
import firedust

assistant = firedust.assistant.create("ASSISTANT_NAME")
response = assistant.chat.message("Hello, how are you?")
print(response.message)
```
"""

from typing import Optional

from firedust._assistant.chat.base import AsyncChat, Chat
from firedust._assistant.interface.base import AsyncInterface, Interface
from firedust._assistant.learning.base import AsyncLearning, Learning
from firedust._assistant.memory.base import AsyncMemory, Memory
from firedust.types import AssistantConfig
from firedust.types.base import INFERENCE_MODEL
from firedust.utils.api import AsyncAPIClient, SyncAPIClient
from firedust.utils.errors import APIError, AssistantError
from firedust.utils.logging import LOG


class Assistant:
    """
    The main class to interact with firedust AI assistants. The class doesn't support direct
    instantiation. To create a new assistant, load an existing one or update it, use the methods
    firedust.assistant.create, firedust.assistant.load, and assistant.update.*.

    Example:
    ```python
    import firedust

    assistant = firedust.assistant.create("ASSISTANT_NAME")
    response = assistant.chat.message("Hello, how are you?")
    print(response.message)

    assistant = firedust.assistant.load("ASSISTANT_NAME")
    assistant.update.instructions("1. Protect the ring bearer. 2. Do not let the ring corrupt you.")
    ```

    Attributes:
        config (AssistantConfig): The assistant configuration.
        api_client (SyncAPIClient): The API client.
        update (Update): Methods to update the configuration of the assistant.
        interface (Interface): Deploy and interact with the assistant on Slack, email, Discord and other interfaces.
        learn (Learning): Methods to add your data to the assistant's knowledge.
        chat (Chat): Methods to chat with the assistant.
        memory (Memory): Methods to interact with the assistant's memory.

    Private Attributes:
        _allow_instantiation (bool): A private attribute to discourage direct instantiation of the Assistant class.
    """

    _config: AssistantConfig
    _api_client: SyncAPIClient

    _update: "_Update"
    _interface: Interface
    _learn: Learning
    _chat: Chat
    _memory: Memory

    # assistant should be started with firedust.assistant.create or .load methods
    _allow_instantiation: bool = False

    def __init__(
        self, config: AssistantConfig, api_client: Optional[SyncAPIClient] = None
    ) -> None:
        # prevent direct instantiation
        # use firedust.assistant.create or .load methods
        if not self._allow_instantiation:
            raise AssistantError(
                """
                To load an assistant or create a new one use firedust.assitant.create or .load methods.
                
                Example:
                ```python
                import firedust

                assistant = firedust.assistant.create("ASSISTANT_NAME")
                # or
                assistant = firedust.assistant.load("ASSISTANT_NAME")

                response = assistant.chat.message("Hello, how are you?")
                print(response.message)
                """
            )

        # configuration
        self._config = config
        self._api_client = api_client or SyncAPIClient()

        # management
        self._update = _Update(self._config, self._api_client)
        self._interface = Interface(self._config, self._api_client)

        # essence
        self._learn = Learning(self._config, self._api_client)
        self._chat = Chat(self._config, self._api_client)
        self._memory = Memory(self._config, self._api_client)

    def __setattr__(self, key: str, value: str) -> None:
        """
        Raise a custom error when trying to set an attribute directly.
        To update the assistant, use the methods assistant.update.* or create a new one.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        # or
        assistant = firedust.assistant.create("ASSISTANT_NAME")

        assistant.update.instructions("1. Protect the ring bearer. 2. Do not let the ring corrupt you.")
        assistant.update.model("mistral/mistral-medium")
        ```
        """
        if not key.startswith("_"):
            self._raise_setter_error(key)
        return super().__setattr__(key, value)

    def __repr__(self) -> str:
        return (
            "Assistant("
            f"name={self.config.name!r}, "
            f"instructions={self.config.instructions!r}, "
            f"model={self.config.model!r}"
            ")"
        )

    def __str__(self) -> str:
        return (
            f"assistant: {self.config.name}, "
            f"model: {self.config.model}, "
            f"instructions: {self.config.instructions}"
        )

    @property
    def config(self) -> AssistantConfig:
        return self._config

    @property
    def api_client(self) -> SyncAPIClient:
        return self._api_client

    @property
    def update(self) -> "_Update":
        return self._update

    @property
    def interface(self) -> Interface:
        return self._interface

    @property
    def learn(self) -> Learning:
        return self._learn

    @property
    def chat(self) -> Chat:
        return self._chat

    @property
    def memory(self) -> Memory:
        return self._memory

    @classmethod
    def _create_instance(
        cls, config: AssistantConfig, api_client: SyncAPIClient
    ) -> "Assistant":
        """
        Reserved for internal use only. To create an assistant use firedust.assistant.create method

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.create("ASSISTANT_NAME")
        response = assistant.chat.message("Hello, how are you?")
        print(response.message)
        ```

        Args:
            config (AssistantConfig): The assistant configuration.
            api_client (SyncAPIClient): The API client.

        Returns:
            Assistant: A new instance of the Assistant class.
        """
        cls._allow_instantiation = True
        assistant = cls(config, api_client)
        cls._allow_instantiation = False
        return assistant

    @staticmethod
    def _raise_setter_error(attribute: str) -> None:
        raise AttributeError(
            f"""
                Unable to set attribute '{attribute}'. Assistant class is immutable and doesn't support direct attribute assignment.
                To update assistant, use assistant.update.* methods or create a new assistant.
                
                Example:
                ```python
                import firedust

                assistant = firedust.assistant.load("ASSISTANT_NAME")
                assistant.update.instructions("1. Protect the ring bearer. 2. Do not let the ring corrupt you.")
                assistant.update.model("mistral/mistral-medium")
                ```
            """
        )

    def delete(
        self,
        confirm: bool = False,
    ) -> None:
        """
        Delete the assistant irreversibly.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        assistant.delete(confirm=True)
        ```

        Args:
            confirm (bool, optional): Confirm the deletion of the assistant. Defaults to False.
        """
        if not confirm:
            raise AssistantError(
                "Assistant and all its memories will be permanently deleted. To confirm, set confirm=True."
            )

        response = self.api_client.delete(
            "/assistant", params={"name": self.config.name}
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"An error occured while deleting the assistant {self.config.name}: {response.text}",
            )
        LOG.info(f"Successfully deleted assistant {self.config.name}.")


class _Update:
    """
    A collection of methods to update an existing assistant.

    Example:
    ```python
    import firedust

    assistant = firedust.assistant.load("ASSISTANT_NAME")

    new_instructions = "1. Protect the ring bearer. 2. Do not let the ring corrupt you."
    new_model = "mistral/mistral-medium"

    assistant.update.instructions(new_instructions)
    assistant.update.model(new_model)
    ```
    """

    def __init__(self, config: AssistantConfig, api_client: SyncAPIClient) -> None:
        self.config = config
        self.api_client = api_client

    def instructions(self, instructions: str) -> None:
        """
        Updates the instructions of the assistant.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")

        new_instructions = "1. Protect the ring bearer. 2. Do not let the ring corrupt you."
        assistant.update.instructions(new_instructions)
        ```

        Args:
            instructions (str): The new instructions for the assistant.
        """
        response = self.api_client.put(
            "/assistant/instructions",
            data={
                "assistant": str(self.config.name),
                "instructions": instructions,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to update the instructions of the assistant {self.config.name}: {response.text}",
            )
        self.config.instructions = instructions

    def model(self, new_model: INFERENCE_MODEL) -> None:
        """
        Updates the inference model of the assistant.

        Example:
        ```python
        import firedust

        assistant = firedust.assistant.load("ASSISTANT_NAME")
        assistant.update.model("mistral/mistral-medium")
        ```

        Args:
            new_model (INFERENCE_MODEL): The new inference model.
        """
        response = self.api_client.put(
            "/assistant/model",
            data={
                "assistant": str(self.config.name),
                "model": new_model,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to update the inference configuration: {response.text}",
            )
        self.config.model = new_model


class AsyncAssistant:
    """
    The main class to interact with firedust AI assistants asynchronously. The class doesn't support direct
    instantiation. To create a new assistant, load an existing one or update it, use the methods
    firedust.assistant.async_create, firedust.assistant.async_load, and assistant.update.*.

    Example:
    ```python
    import firedust
    import asyncio

    async def main():
        assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
        response = await assistant.chat.message("Hello, how are you?")
        print(response.message)

    asyncio.run(main())
    ```

    Attributes:
        config (AssistantConfig): The assistant configuration.
        api_client (AsyncAPIClient): The API client.
        update (_AsyncUpdate): Methods to update the configuration of the assistant.
        interface (AsyncInterface): Deploy and interact with the assistant on Slack, email, Discord and other interfaces.
        learn (AsyncLearning): Methods to train the assistant on your data.
        chat (AsyncChat): Methods to chat with the assistant.
        memory (AsyncMemory): Methods to interact with the assistant's memory.

    Private Attributes:
        _allow_instantiation (bool): A private attribute to discourage direct instantiation of the AsyncAssistant class.
    """

    _config: AssistantConfig
    _api_client: AsyncAPIClient

    _update: "_AsyncUpdate"
    _interface: AsyncInterface
    _learn: AsyncLearning
    _chat: AsyncChat
    _memory: AsyncMemory

    # assistant should be instantiated using the create and load classmethods
    _allow_instantiation: bool = False

    def __init__(
        self, config: AssistantConfig, api_client: Optional[AsyncAPIClient] = None
    ) -> None:
        # prevent direct instantiation
        if not self._allow_instantiation:
            raise AssistantError(
                """
                To load an assistant or create a new one use firedust.assitant.async_create or .async_load methods.
                
                Example:
                ```python
                import firedust
                import asyncio

                async def main():
                    assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
                    # or
                    assistant = await firedust.assistant.async_create("ASSISTANT_NAME")
                    
                    response = await assistant.chat.message("Hello, how are you?")
                    print(response.message)

                asyncio.run(main())
                """
            )

        # configuration
        self._config = config
        self._api_client = api_client or AsyncAPIClient()

        # management
        self._update = _AsyncUpdate(self._config, self._api_client)
        self._interface = AsyncInterface(self._config, self._api_client)

        # essence
        self._learn = AsyncLearning(self._config, self._api_client)
        self._chat = AsyncChat(self._config, self._api_client)
        self._memory = AsyncMemory(self._config, self._api_client)

    def __setattr__(self, key: str, value: str) -> None:
        """
        Raise a custom error when trying to set an attribute directly.
        To update the assistant, use the methods assistant.update.* or create a new one.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
            # or
            assistant = await firedust.assistant.async_create("ASSISTANT_NAME")

            await assistant.update.instructions("1. Protect the ring bearer. 2. Do not let the ring corrupt you.")
            await assistant.update.model("mistral/mistral-medium")

        asyncio.run(main())
        ```
        """
        if not key.startswith("_"):
            self._raise_setter_error(key)
        return super().__setattr__(key, value)

    def __repr__(self) -> str:
        return (
            "AsyncAssistant("
            f"name={self.config.name!r}, "
            f"instructions={self.config.instructions!r}, "
            f"model={self.config.model!r}"
            ")"
        )

    def __str__(self) -> str:
        return (
            f"assistant: {self.config.name}, "
            f"model: {self.config.model}, "
            f"instructions: {self.config.instructions}"
        )

    @property
    def config(self) -> AssistantConfig:
        return self._config

    @property
    def api_client(self) -> AsyncAPIClient:
        return self._api_client

    @property
    def update(self) -> "_AsyncUpdate":
        return self._update

    @property
    def interface(self) -> AsyncInterface:
        return self._interface

    @property
    def learn(self) -> AsyncLearning:
        return self._learn

    @property
    def chat(self) -> AsyncChat:
        return self._chat

    @property
    def memory(self) -> AsyncMemory:
        return self._memory

    @classmethod
    async def _create_instance(
        cls, config: AssistantConfig, api_client: AsyncAPIClient
    ) -> "AsyncAssistant":
        """
        Reserved for internal use only. To create an assistant use firedust.assistant.async_create method.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_create("ASSISTANT_NAME")
            response = await assistant.chat.message("Hello, how are you?")
            print(response.message)

        asyncio.run(main())
        ```
        Args:
            config (AssistantConfig): The assistant configuration.
            api_client (AsyncAPIClient): The API client.

        Returns:
            AsyncAssistant: A new instance of the AsyncAssistant class.
        """
        cls._allow_instantiation = True
        assistant = cls(config, api_client)
        cls._allow_instantiation = False
        return assistant

    @staticmethod
    def _raise_setter_error(attribute: str) -> None:
        raise AttributeError(
            f"""
                Unable to set attribute '{attribute}'. AsyncAssistant class is immutable and doesn't support direct attribute assignment.
                To update assistant, use assistant.update.* methods or create a new assistant.

                Example:
                ```python
                import firedust
                import asyncio

                async def main():
                    assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
                    # or
                    assistant = await firedust.assistant.async_create("ASSISTANT_NAME")

                    await assistant.update.instructions("1. Protect the ring bearer. 2. Do not let the ring corrupt you.")
                    await assistant.update.model("mistral/mistral-medium")

                asyncio.run(main())
            """
        )

    async def delete(
        self,
        confirm: bool = False,
    ) -> None:
        """
        Deletes an existing assistant irreversibly.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
            await assistant.delete(confirm=True)

        asyncio.run(main())
        ```

        Args:
            confirm (bool, optional): Confirm the deletion of the assistant. Defaults to False.
        """
        if not confirm:
            raise AssistantError(
                f"Assistant {self.config.name} and all its memories will be permanently deleted. To confirm, set confirm=True."
            )

        response = await self.api_client.delete(
            "/assistant", params={"name": self.config.name}
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"An error occured while deleting the assistant {self.config.name}: {response.text}",
            )
        LOG.info(f"Successfully deleted assistant {self.config.name}.")


class _AsyncUpdate:
    """
    A collection of methods to update an existing assistant asynchronously.

    Example:
    ```python
    import firedust
    import asyncio

    async def main():
        assistant = await firedust.assistant.async_load("ASSISTANT_NAME")

        new_instructions = "1. Protect the ring bearer. 2. Do not let the ring corrupt you."
        new_model = "mistral/mistral-medium"

        await assistant.update.instructions(new_instructions)
        await assistant.update.model(new_model)

    asyncio.run(main())
    ```
    """

    def __init__(self, config: AssistantConfig, api_client: AsyncAPIClient) -> None:
        self.config = config
        self.api_client = api_client

    async def instructions(self, instructions: str) -> None:
        """
        Updates the instructions of the assistant.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")

            new_instructions = "1. Protect the ring bearer. 2. Do not let the ring corrupt you."
            await assistant.update.instructions(new_instructions)

        asyncio.run(main())
        ```

        Args:
            instructions (str): The new instructions of the assistant.
        """
        response = await self.api_client.put(
            "/assistant/instructions",
            data={
                "assistant": str(self.config.name),
                "instructions": instructions,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to update the instructions of the assistant {self.config.name}: {response.text}",
            )
        self.config = AssistantConfig(
            name=self.config.name,
            instructions=instructions,
            model=self.config.model,
        )

    async def model(self, new_model: INFERENCE_MODEL) -> None:
        """
        Updates the inference model of the assistant.

        Example:
        ```python
        import firedust
        import asyncio

        async def main():
            assistant = await firedust.assistant.async_load("ASSISTANT_NAME")
            await assistant.update.model("mistral/mistral-medium")

        asyncio.run(main())
        ```

        Args:
            new_model (INFERENCE_MODEL): The new inference model.
        """
        response = await self.api_client.put(
            "/assistant/model",
            data={
                "assistant": str(self.config.name),
                "model": new_model,
            },
        )
        if not response.is_success:
            raise APIError(
                code=response.status_code,
                message=f"Failed to update the inference configuration: {response.text}",
            )
        self.config = AssistantConfig(
            name=self.config.name,
            instructions=self.config.instructions,
            model=new_model,
        )
